# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""BigEarthNet dataset."""

import glob
import json
import os
import textwrap
from collections.abc import Callable
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
import torch
from matplotlib.colors import BoundaryNorm, ListedColormap
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from rasterio.enums import Resampling
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_url, extract_archive, sort_sentinel2_bands


class BigEarthNet(NonGeoDataset):
    """BigEarthNet dataset.

    The `BigEarthNet <https://bigearth.net/>`__
    dataset is a dataset for multilabel remote sensing image scene classification.

    Dataset features:

    * 590,326 patches from 125 Sentinel-1 and Sentinel-2 tiles
    * Imagery from tiles in Europe between Jun 2017 - May 2018
    * 12 spectral bands with 10-60 m per pixel resolution (base 120x120 px)
    * 2 synthetic aperture radar bands (120x120 px)
    * 43 or 19 scene classes from the 2018 CORINE Land Cover database (CLC 2018)

    Dataset format:

    * images are composed of multiple single channel geotiffs
    * labels are multiclass, stored in a single json file per image
    * mapping of Sentinel-1 to Sentinel-2 patches are within Sentinel-1 json files
    * Sentinel-1 bands: (VV, VH)
    * Sentinel-2 bands: (B01, B02, B03, B04, B05, B06, B07, B08, B8A, B09, B11, B12)
    * All bands: (VV, VH, B01, B02, B03, B04, B05, B06, B07, B08, B8A, B09, B11, B12)
    * Sentinel-2 bands are of different spatial resolutions and upsampled to 10m

    Dataset classes (43):

    0. Continuous urban fabric
    1. Discontinuous urban fabric
    2. Industrial or commercial units
    3. Road and rail networks and associated land
    4. Port areas
    5. Airports
    6. Mineral extraction sites
    7. Dump sites
    8. Construction sites
    9. Green urban areas
    10. Sport and leisure facilities
    11. Non-irrigated arable land
    12. Permanently irrigated land
    13. Rice fields
    14. Vineyards
    15. Fruit trees and berry plantations
    16. Olive groves
    17. Pastures
    18. Annual crops associated with permanent crops
    19. Complex cultivation patterns
    20. Land principally occupied by agriculture, with significant
        areas of natural vegetation
    21. Agro-forestry areas
    22. Broad-leaved forest
    23. Coniferous forest
    24. Mixed forest
    25. Natural grassland
    26. Moors and heathland
    27. Sclerophyllous vegetation
    28. Transitional woodland/shrub
    29. Beaches, dunes, sands
    30. Bare rock
    31. Sparsely vegetated areas
    32. Burnt areas
    33. Inland marshes
    34. Peatbogs
    35. Salt marshes
    36. Salines
    37. Intertidal flats
    38. Water courses
    39. Water bodies
    40. Coastal lagoons
    41. Estuaries
    42. Sea and ocean

    Dataset classes (19):

    0. Urban fabric
    1. Industrial or commercial units
    2. Arable land
    3. Permanent crops
    4. Pastures
    5. Complex cultivation patterns
    6. Land principally occupied by agriculture, with significant
       areas of natural vegetation
    7. Agro-forestry areas
    8. Broad-leaved forest
    9. Coniferous forest
    10. Mixed forest
    11. Natural grassland and sparsely vegetated areas
    12. Moors, heathland and sclerophyllous vegetation
    13. Transitional woodland, shrub
    14. Beaches, dunes, sands
    15. Inland wetlands
    16. Coastal wetlands
    17. Inland waters
    18. Marine waters

    The source for the above dataset classes, their respective ordering, and
    43-to-19-class mappings can be found here:

    * https://git.tu-berlin.de/rsim/BigEarthNet-S2_19-classes_models/-/blob/master/label_indices.json

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1109/IGARSS.2019.8900532

    """

    class_sets: ClassVar[dict[int, list[str]]] = {
        19: [
            'Urban fabric',
            'Industrial or commercial units',
            'Arable land',
            'Permanent crops',
            'Pastures',
            'Complex cultivation patterns',
            'Land principally occupied by agriculture, with significant areas of'
            ' natural vegetation',
            'Agro-forestry areas',
            'Broad-leaved forest',
            'Coniferous forest',
            'Mixed forest',
            'Natural grassland and sparsely vegetated areas',
            'Moors, heathland and sclerophyllous vegetation',
            'Transitional woodland, shrub',
            'Beaches, dunes, sands',
            'Inland wetlands',
            'Coastal wetlands',
            'Inland waters',
            'Marine waters',
        ],
        43: [
            'Continuous urban fabric',
            'Discontinuous urban fabric',
            'Industrial or commercial units',
            'Road and rail networks and associated land',
            'Port areas',
            'Airports',
            'Mineral extraction sites',
            'Dump sites',
            'Construction sites',
            'Green urban areas',
            'Sport and leisure facilities',
            'Non-irrigated arable land',
            'Permanently irrigated land',
            'Rice fields',
            'Vineyards',
            'Fruit trees and berry plantations',
            'Olive groves',
            'Pastures',
            'Annual crops associated with permanent crops',
            'Complex cultivation patterns',
            'Land principally occupied by agriculture, with significant areas of'
            ' natural vegetation',
            'Agro-forestry areas',
            'Broad-leaved forest',
            'Coniferous forest',
            'Mixed forest',
            'Natural grassland',
            'Moors and heathland',
            'Sclerophyllous vegetation',
            'Transitional woodland/shrub',
            'Beaches, dunes, sands',
            'Bare rock',
            'Sparsely vegetated areas',
            'Burnt areas',
            'Inland marshes',
            'Peatbogs',
            'Salt marshes',
            'Salines',
            'Intertidal flats',
            'Water courses',
            'Water bodies',
            'Coastal lagoons',
            'Estuaries',
            'Sea and ocean',
        ],
    }

    label_converter: ClassVar[dict[int, int]] = {
        0: 0,
        1: 0,
        2: 1,
        11: 2,
        12: 2,
        13: 2,
        14: 3,
        15: 3,
        16: 3,
        18: 3,
        17: 4,
        19: 5,
        20: 6,
        21: 7,
        22: 8,
        23: 9,
        24: 10,
        25: 11,
        31: 11,
        26: 12,
        27: 12,
        28: 13,
        29: 14,
        33: 15,
        34: 15,
        35: 16,
        36: 16,
        38: 17,
        39: 17,
        40: 18,
        41: 18,
        42: 18,
    }

    splits_metadata: ClassVar[dict[str, dict[str, str]]] = {
        'train': {
            'url': 'https://git.tu-berlin.de/rsim/BigEarthNet-MM_19-classes_models/-/raw/9a5be07346ab0884b2d9517475c27ef9db9b5104/splits/train.csv?inline=false',
            'filename': 'bigearthnet-train.csv',
            'md5': '623e501b38ab7b12fe44f0083c00986d',
        },
        'val': {
            'url': 'https://git.tu-berlin.de/rsim/BigEarthNet-MM_19-classes_models/-/raw/9a5be07346ab0884b2d9517475c27ef9db9b5104/splits/val.csv?inline=false',
            'filename': 'bigearthnet-val.csv',
            'md5': '22efe8ed9cbd71fa10742ff7df2b7978',
        },
        'test': {
            'url': 'https://git.tu-berlin.de/rsim/BigEarthNet-MM_19-classes_models/-/raw/9a5be07346ab0884b2d9517475c27ef9db9b5104/splits/test.csv?inline=false',
            'filename': 'bigearthnet-test.csv',
            'md5': '697fb90677e30571b9ac7699b7e5b432',
        },
    }
    metadata: ClassVar[dict[str, dict[str, str]]] = {
        's1': {
            'url': 'https://zenodo.org/records/12687186/files/BigEarthNet-S1-v1.0.tar.gz',
            'md5': '94ced73440dea8c7b9645ee738c5a172',
            'filename': 'BigEarthNet-S1-v1.0.tar.gz',
            'directory': 'BigEarthNet-S1-v1.0',
        },
        's2': {
            'url': 'https://zenodo.org/records/12687186/files/BigEarthNet-S2-v1.0.tar.gz',
            'md5': '5a64e9ce38deb036a435a7b59494924c',
            'filename': 'BigEarthNet-S2-v1.0.tar.gz',
            'directory': 'BigEarthNet-v1.0',
        },
    }
    image_size = (120, 120)

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        bands: str = 'all',
        num_classes: int = 19,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new BigEarthNet dataset instance.

        Args:
            root: root directory where dataset can be found
            split: train/val/test split to load
            bands: load Sentinel-1 bands, Sentinel-2, or both. one of {s1, s2, all}
            num_classes: number of classes to load in target. one of {19, 43}
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.splits_metadata
        assert bands in ['s1', 's2', 'all']
        assert num_classes in [43, 19]
        self.root = root
        self.split = split
        self.bands = bands
        self.num_classes = num_classes
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self.class2idx = {c: i for i, c in enumerate(self.class_sets[43])}
        self._verify()
        self.folders = self._load_folders()

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        image = self._load_image(index)
        label = self._load_target(index)
        sample: dict[str, Tensor] = {'image': image, 'label': label}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.folders)

    def _load_folders(self) -> list[dict[str, str]]:
        """Load folder paths.

        Returns:
            list of dicts of s1 and s2 folder paths
        """
        filename = self.splits_metadata[self.split]['filename']
        dir_s1 = self.metadata['s1']['directory']
        dir_s2 = self.metadata['s2']['directory']

        with open(os.path.join(self.root, filename)) as f:
            lines = f.read().strip().splitlines()
            pairs = [line.split(',') for line in lines]

        folders = [
            {
                's1': os.path.join(self.root, dir_s1, pair[1]),
                's2': os.path.join(self.root, dir_s2, pair[0]),
            }
            for pair in pairs
        ]
        return folders

    def _load_paths(self, index: int) -> list[str]:
        """Load paths to band files.

        Args:
            index: index to return

        Returns:
            list of file paths
        """
        if self.bands == 'all':
            folder_s1 = self.folders[index]['s1']
            folder_s2 = self.folders[index]['s2']
            paths_s1 = glob.glob(os.path.join(folder_s1, '*.tif'))
            paths_s2 = glob.glob(os.path.join(folder_s2, '*.tif'))
            paths_s1 = sorted(paths_s1)
            paths_s2 = sorted(paths_s2, key=sort_sentinel2_bands)
            paths = paths_s1 + paths_s2
        elif self.bands == 's1':
            folder = self.folders[index]['s1']
            paths = glob.glob(os.path.join(folder, '*.tif'))
            paths = sorted(paths)
        else:
            folder = self.folders[index]['s2']
            paths = glob.glob(os.path.join(folder, '*.tif'))
            paths = sorted(paths, key=sort_sentinel2_bands)

        return paths

    def _load_image(self, index: int) -> Tensor:
        """Load a single image.

        Args:
            index: index to return

        Returns:
            the raster image or target
        """
        paths = self._load_paths(index)
        images = []
        for path in paths:
            # Bands are of different spatial resolutions
            # Resample to (120, 120)
            with rasterio.open(path) as dataset:
                array = dataset.read(
                    indexes=1,
                    out_shape=self.image_size,
                    out_dtype='int32',
                    resampling=Resampling.bilinear,
                )
                images.append(array)
        arrays: np.typing.NDArray[np.int_] = np.stack(images, axis=0)
        tensor = torch.from_numpy(arrays).float()
        return tensor

    def _load_target(self, index: int) -> Tensor:
        """Load the target mask for a single image.

        Args:
            index: index to return

        Returns:
            the target label
        """
        if self.bands == 's2':
            folder = self.folders[index]['s2']
        else:
            folder = self.folders[index]['s1']

        path = glob.glob(os.path.join(folder, '*.json'))[0]
        with open(path) as f:
            labels = json.load(f)['labels']

        # labels -> indices
        indices = [self.class2idx[label] for label in labels]

        # Map 43 to 19 class labels
        if self.num_classes == 19:
            indices_optional = [self.label_converter.get(idx) for idx in indices]
            indices = [idx for idx in indices_optional if idx is not None]

        target = torch.zeros(self.num_classes, dtype=torch.long)
        target[indices] = 1
        return target

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        keys = ['s1', 's2'] if self.bands == 'all' else [self.bands]
        urls = [self.metadata[k]['url'] for k in keys]
        md5s = [self.metadata[k]['md5'] for k in keys]
        filenames = [self.metadata[k]['filename'] for k in keys]
        directories = [self.metadata[k]['directory'] for k in keys]
        urls.extend([self.splits_metadata[k]['url'] for k in self.splits_metadata])
        md5s.extend([self.splits_metadata[k]['md5'] for k in self.splits_metadata])
        filenames_splits = [
            self.splits_metadata[k]['filename'] for k in self.splits_metadata
        ]
        filenames.extend(filenames_splits)

        # Check if the split file already exist
        exists = []
        for filename in filenames_splits:
            exists.append(os.path.exists(os.path.join(self.root, filename)))

        # Check if the files already exist
        for directory in directories:
            exists.append(os.path.exists(os.path.join(self.root, directory)))

        if all(exists):
            return

        # Check if zip file already exists (if so then extract)
        exists = []
        for filename in filenames:
            filepath = os.path.join(self.root, filename)
            if os.path.exists(filepath):
                exists.append(True)
                self._extract(filepath)
            else:
                exists.append(False)

        if all(exists):
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download and extract the dataset
        for url, filename, md5 in zip(urls, filenames, md5s):
            self._download(url, filename, md5)
            filepath = os.path.join(self.root, filename)
            self._extract(filepath)

    def _download(self, url: str, filename: Path, md5: str) -> None:
        """Download the dataset.

        Args:
            url: url to download file
            filename: output filename to write downloaded file
            md5: md5 of downloaded file
        """
        download_url(
            url, self.root, filename=filename, md5=md5 if self.checksum else None
        )

    def _extract(self, filepath: Path) -> None:
        """Extract the dataset.

        Args:
            filepath: path to file to be extracted
        """
        if not str(filepath).endswith('.csv'):
            extract_archive(filepath)

    def _onehot_labels_to_names(
        self, label_mask: 'np.typing.NDArray[np.bool_]'
    ) -> list[str]:
        """Gets a list of class names given a label mask.

        Args:
            label_mask: a boolean mask corresponding to a set of labels or predictions

        Returns:
            a list of class names corresponding to the input mask
        """
        labels = []
        for i, mask in enumerate(label_mask):
            if mask:
                labels.append(self.class_sets[self.num_classes][i])
        return labels

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        .. versionadded:: 0.2
        """
        if self.bands == 's2':
            image = np.rollaxis(sample['image'][[3, 2, 1]].numpy(), 0, 3)
            image = np.clip(image / 2000, 0, 1)
        elif self.bands == 'all':
            image = np.rollaxis(sample['image'][[5, 4, 3]].numpy(), 0, 3)
            image = np.clip(image / 2000, 0, 1)
        elif self.bands == 's1':
            image = sample['image'][0].numpy()

        label_mask = sample['label'].numpy().astype(np.bool_)
        labels = self._onehot_labels_to_names(label_mask)

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction_mask = sample['prediction'].numpy().astype(np.bool_)
            predictions = self._onehot_labels_to_names(prediction_mask)

        fig, ax = plt.subplots(figsize=(4, 4))
        ax.imshow(image)
        ax.axis('off')
        if show_titles:
            title = f'Labels: {", ".join(labels)}'
            if showing_predictions:
                title += f'\nPredictions: {", ".join(predictions)}'
            ax.set_title(title)

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig


class BigEarthNetV2(NonGeoDataset):
    """BigEarthNetV2 dataset.

    The `BigEarthNet V2 <https://bigearth.net/>`__ dataset contains improved labels, improved
    geospatial data splits and additionally pixel-level labels from CORINE Land
    Cover (CLC) map of 2018. Additionally, some problematic patches from V1 have been removed.

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2407.03653

    .. versionadded:: 0.7
    """

    class_set = BigEarthNet.class_sets[19]

    image_size = BigEarthNet.image_size

    url = 'https://hf.co/datasets/torchgeo/bigearthnet/resolve/3cf3a5910a5302d449fdb8e570e5b78de24fe07f/V2/{}'

    metadata_locs: ClassVar[dict[str, dict[str, dict[str, str]]]] = {
        's1': {
            'files': {
                'BigEarthNet-S1.tar.gzaa': '039b9ce305fc6582b2c3d60d1573f5b7',
                'BigEarthNet-S1.tar.gzab': 'e94f0ea165d04992ca91d8e58e82ec6d',
            }
        },
        's2': {
            'files': {
                'BigEarthNet-S2.tar.gzaa': '94e8ed32065234d3ab46353d814778d1',
                'BigEarthNet-S2.tar.gzab': '24c223d9e36166136c13b24a27debe34',
            }
        },
        'maps': {
            'files': {'Reference_Maps.tar.gzaa': 'b0cd1f0a31b49fcbfd61d80f963e759d'}
        },
        'metadata': {'files': {'metadata.parquet': '55687065e77b6d0b0f1ff604a6e7b49c'}},
    }

    dir_file_names: ClassVar[dict[str, str]] = {
        's1': 'BigEarthNet-S1',
        's2': 'BigEarthNet-S2',
        'maps': 'Reference_Maps',
        'metadata': 'metadata.parquet',
    }

    # https://collections.sentinel-hub.com/corine-land-cover/readme.html
    # Table 1 of https://bigearth.net/static/documents/Description_BigEarthNet_v2.pdf
    clc_colors: ClassVar[dict[str, str]] = {
        'Urban fabric': '#e6004d',
        'Industrial or commercial units': '#cc4df2',
        'Arable land': '#ffffa8',
        'Permanent crops': '#e68000',
        'Pastures': '#e6e64d',
        'Complex cultivation patterns': '#ffe64d',
        'Land principally occupied by agriculture, with significant areas of natural vegetation': '#e6cc4d',
        'Agro-forestry areas': '#f2cca6',
        'Broad-leaved forest': '#80ff00',
        'Coniferous forest': '#00a600',
        'Mixed forest': '#4dff00',
        'Natural grassland and sparsely vegetated areas': '#ccf24d',
        'Moors, heathland and sclerophyllous vegetation': '#a6ff80',
        'Transitional woodland, shrub': '#a6f200',
        'Beaches, dunes, sands': '#e6e6e6',
        'Inland wetlands': '#a6a6ff',
        'Coastal wetlands': '#ccccff',
        'Inland waters': '#80f2e6',
        'Marine waters': '#e6f2ff',
    }

    clc_codes: ClassVar[dict[int, int]] = {
        111: 0,  # Continuous Urban fabric
        112: 0,  # Discontinuous Urban fabric
        121: 1,  # Industrial or commercial units
        211: 2,  # Non-irrigated arable land
        212: 2,  # Permanently irrigated land
        213: 2,  # Rice fields
        221: 3,  # Vineyards
        222: 3,  # Fruit trees and berry plantations
        223: 3,  # Olive groves
        231: 4,  # Pastures
        241: 3,  # Annual crops with permanent crops
        242: 5,  # Complex cultivation patterns
        243: 6,  # Land principally occupied by agriculture...
        244: 7,  # Agro-forestry areas
        311: 8,  # Broad-leaved forest
        312: 9,  # Coniferous forest
        313: 10,  # Mixed forest
        321: 11,  # Natural grassland
        322: 12,  # Moors and heathland
        323: 12,  # Sclerophyllous vegetation
        324: 13,  # Transitional woodland/shrub
        331: 14,  # Beaches, dunes, sands
        333: 11,  # Sparsely vegetated areas
        411: 15,  # Inland marshes
        412: 15,  # Peatbogs
        421: 16,  # Salt marshes
        422: 16,  # Salines
        511: 17,  # Water courses
        512: 17,  # Water bodies
        521: 18,  # Coastal lagoons
        522: 18,  # Estuaries
        523: 18,  # Sea and ocean
    }

    valid_splits = ('train', 'val', 'test')

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        bands: str = 'all',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new BigEarthNet V2 dataset instance.

        Args:
            root: root directory where dataset can be found
            split: train/val/test split to load
            bands: load Sentinel-1 bands, Sentinel-2, or both. one of {s1, s2, all}
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
            AssertionError: If *split*, or *bands*, are not valid.
        """
        assert split in self.valid_splits, f'split must be one of {self.valid_splits}'
        assert bands in ['s1', 's2', 'all']
        self.root = root
        self.split = split
        self.bands = bands
        self.transforms = transforms
        self.num_classes = 19
        self.download = download
        self.checksum = checksum
        self.class2idx = {c: i for i, c in enumerate(self.class_set)}
        self._verify()

        self.metadata_df = pd.read_parquet(os.path.join(self.root, 'metadata.parquet'))
        self.metadata_df = self.metadata_df[
            self.metadata_df['split'] == self.split
        ].reset_index(drop=True)

        # Map chosen classes to ordinal numbers, all others mapped to background class
        self.ordinal_map = torch.zeros(19)
        for corine, ordinal in self.clc_codes.items():
            self.ordinal_map[ordinal] = corine

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.metadata_df)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        sample: dict[str, Tensor] = {}

        match self.bands:
            case 's1':
                sample['image'] = self._load_image(index, 's1')
            case 's2':
                sample['image'] = self._load_image(index, 's2')
            case 'all':
                sample['image_s1'] = self._load_image(index, 's1')
                sample['image_s2'] = self._load_image(index, 's2')

        sample['mask'] = self._load_map(index)
        sample['label'] = self._load_target(index)

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _load_image(self, index: int, sensor: str) -> Tensor:
        """Generic image loader for both S1 and S2.

        Args:
            index: index to return
            sensor: 's1' or 's2'

        Returns:
            the sensor image
        """
        row = self.metadata_df.loc[index]
        id_field = 's1_name' if sensor == 's1' else 'patch_id'
        patch_id = row[id_field]
        if sensor == 's2':
            patch_dir = '_'.join(patch_id.split('_')[0:-2])
        else:
            patch_dir = '_'.join(patch_id.split('_')[0:-3])

        paths = glob.glob(
            os.path.join(
                self.root, self.dir_file_names[sensor], patch_dir, patch_id, '*.tif'
            )
        )

        if sensor == 's2':
            paths = sorted(paths, key=sort_sentinel2_bands)
        else:
            paths = sorted(paths)

        images = []
        for path in paths:
            with rasterio.open(path) as dataset:
                array = dataset.read(
                    indexes=1,
                    out_shape=self.image_size,
                    out_dtype='int32',
                    resampling=Resampling.bilinear,
                )
                images.append(array)

        return torch.from_numpy(np.stack(images, axis=0)).float()

    def _load_map(self, index: int) -> Tensor:
        """Load a single image.

        Args:
            index: index to return

        Returns:
            the Corine Land Cover map
        """
        row = self.metadata_df.loc[index]
        patch_id = row['patch_id']
        patch_dir = '_'.join(patch_id.split('_')[0:-2])
        path = os.path.join(
            self.root,
            self.dir_file_names['maps'],
            patch_dir,
            patch_id,
            patch_id + '_reference_map.tif',
        )
        with rasterio.open(path) as dataset:
            map = dataset.read(out_dtype='int32')

        tensor = torch.from_numpy(map)
        # remap to ordinal values
        for corine, ordinal in self.clc_codes.items():
            tensor[tensor == corine] = ordinal
        return tensor.long()

    def _load_target(self, index: int) -> Tensor:
        """Load the target mask for a single image.

        Args:
            index: index to return

        Returns:
            the target label
        """
        label_names = self.metadata_df.iloc[index]['labels']

        indices = [self.class2idx[label_names] for label_names in label_names]

        image_target = torch.zeros(self.num_classes, dtype=torch.long)
        image_target[indices] = 1
        return image_target

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        exists = []
        for key, metadata in self.metadata_locs.items():
            exists.append(
                os.path.exists(os.path.join(self.root, self.dir_file_names[key]))
            )

        if all(exists):
            return

        # check if compressed files already exist
        exists = []
        for key, metadata in self.metadata_locs.items():
            if key == 'metadata':
                exists.append(
                    os.path.exists(os.path.join(self.root, self.dir_file_names[key]))
                )
            else:
                for fname in metadata['files']:
                    fpath = os.path.join(self.root, fname)
                    exists.append(os.path.exists(fpath))

        if all(exists):
            return

        if not self.download:
            raise DatasetNotFoundError(self)

        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the required tarball parts using the URL template and md5 sums."""
        for key, meta in self.metadata_locs.items():
            for fname, md5 in meta['files'].items():
                target_path = os.path.join(self.root, fname)
                if not os.path.exists(target_path):
                    download_url(self.url.format(fname), self.root, md5)

    def _extract(self) -> None:
        """Extract the tarball parts.

        For each modality (s1, s2, maps), its parts are concatenated together and then extracted.
        """
        chunk_size = 2**15  # same as used in torchvision and ssl4eo
        for key, meta in self.metadata_locs.items():
            if key == 'metadata':
                continue
            parts = [os.path.join(self.root, f) for f in meta['files'].keys()]
            concat_path = os.path.join(self.root, self.dir_file_names[key] + '.tar.gz')
            with open(concat_path, 'wb') as outfile:
                for part in parts:
                    with open(part, 'rb') as g:
                        while chunk := g.read(chunk_size):
                            outfile.write(chunk)
            extract_archive(concat_path, self.root)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        fig, axes = plt.subplots(1, 2 if self.bands != 'all' else 3, figsize=(12, 4))

        if self.bands in ['s2', 'all']:
            s2_img = sample['image_s2' if self.bands == 'all' else 'image']
            rgb = np.rollaxis(s2_img[[3, 2, 1]].numpy(), 0, 3)
            axes[0].imshow(np.clip(rgb / 2000, 0, 1))
            if show_titles:
                axes[0].set_title('Sentinel-2 RGB')
            axes[0].axis('off')

        if self.bands in ['s1', 'all']:
            idx = 0 if self.bands == 's1' else 1
            s1_img = sample['image_s1' if self.bands == 'all' else 'image']
            axes[idx].imshow(s1_img[0].numpy())
            if show_titles:
                axes[idx].set_title('Sentinel-1 VV')
            axes[idx].axis('off')

        # Handle mask plotting
        mask_idx = 1 if self.bands != 'all' else 2
        mask = sample['mask'][0].numpy()

        # Get unique ordinal labels from mask
        unique_labels = sorted(np.unique(mask))

        # Map ordinal labels to class names and colors directly
        colors = []
        class_names = []
        for label in unique_labels:
            name = self.class_set[label]  # Get class name from ordinal index
            colors.append(self.clc_colors[name])  # Get color for class name
            class_names.append(name)

        # Create custom colormap
        cmap = ListedColormap(colors)
        bounds = [*unique_labels, unique_labels[-1] + 1]
        norm = BoundaryNorm(bounds, len(colors))

        axes[mask_idx].imshow(mask, cmap=cmap, norm=norm)

        # Add legend with class names
        legend_elements = [Rectangle((0, 0), 1, 1, facecolor=color) for color in colors]
        wrapped_names = [textwrap.fill(name, width=25) for name in class_names]
        axes[mask_idx].legend(
            legend_elements,
            wrapped_names,
            loc='center left',
            bbox_to_anchor=(1, 0.5),
            fontsize='x-small',
        )
        axes[mask_idx].axis('off')

        if show_titles:
            axes[mask_idx].set_title('Land Cover Map')

        if 'label' in sample:
            label_indices = sample['label'].nonzero().squeeze(1).tolist()
            label_names = [self.class_set[idx] for idx in label_indices]
            if suptitle:
                suptitle = f'{suptitle}\nLabels: {", ".join(label_names)}'
            else:
                suptitle = f'Labels: {", ".join(label_names)}'

        if suptitle:
            plt.suptitle(suptitle)

        # Adjust layout to prevent overlap
        plt.tight_layout()

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""xView2 datamodule."""

from typing import Any

import torch
from torch.utils.data import random_split

from ..datasets import XView2
from .geo import NonGeoDataModule


class XView2DataModule(NonGeoDataModule):
    """LightningDataModule implementation for the xView2 dataset.

    Uses the train/val/test splits from the dataset.

    .. versionadded:: 0.2
    """

    def __init__(
        self,
        batch_size: int = 64,
        num_workers: int = 0,
        val_split_pct: float = 0.2,
        **kwargs: Any,
    ) -> None:
        """Initialize a new XView2DataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            val_split_pct: What percentage of the dataset to use as a validation set
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.XView2`.
        """
        super().__init__(XView2, batch_size, num_workers, **kwargs)

        self.val_split_pct = val_split_pct

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        if stage in ['fit', 'validate']:
            self.dataset = XView2(split='train', **self.kwargs)
            generator = torch.Generator().manual_seed(0)
            self.train_dataset, self.val_dataset = random_split(
                self.dataset, [1 - self.val_split_pct, self.val_split_pct], generator
            )
        if stage in ['test']:
            self.test_dataset = XView2(split='test', **self.kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Copernicus-Bench DFC2020-S2 dataset."""

import os

from matplotlib.colors import ListedColormap
from torch import Tensor

from .base import CopernicusBenchBase


class CopernicusBenchDFC2020S2(CopernicusBenchBase):
    """Copernicus-Bench DFC2020-S2 dataset.

    DFC2020-S2 is a land use/land cover segmentation datasets derived from the
    IEEE GRSS Data Fusion Contest 2020 (DFC2020).

    If you use this dataset in your research, please cite the following papers:

    * https://arxiv.org/abs/2503.11849
    * https://ieee-dataport.org/competitions/2020-ieee-grss-data-fusion-contest

    .. versionadded:: 0.7
    """

    url = 'https://hf.co/datasets/wangyi111/Copernicus-Bench/resolve/9d252acd3aa0e3da3128e05c6f028647f0e48e5f/l2_dfc2020_s1s2/dfc2020.zip'
    md5 = 'f10ba017dab6f38b7a6857b169ea924b'
    zipfile = 'dfc2020.zip'
    directory = 'dfc2020_s1s2'
    filename = 'dfc-{}-new.csv'
    all_bands = (
        'B01',
        'B02',
        'B03',
        'B04',
        'B05',
        'B06',
        'B07',
        'B08',
        'B8A',
        'B09',
        'B10',
        'B11',
        'B12',
    )
    rgb_bands = ('B04', 'B03', 'B02')
    classes = (
        'Background',
        'Forest',
        'Shrubland',
        'Savanna',
        'Grassland',
        'Wetlands',
        'Croplands',
        'Urban/Built-up',
        'Snow/Ice',
        'Barren',
        'Water',
    )
    cmap = ListedColormap(
        [
            '#000000',
            '#009900',
            '#c6b044',
            '#fbff13',
            '#b6ff05',
            '#27ff87',
            '#c24f44',
            '#a5a5a5',
            '#69fff8',
            '#f9ffa4',
            '#1c0dff',
        ]
    )

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and labels at that index.
        """
        file = self.files[index]
        image_path = os.path.join(
            self.root, self.directory, 's2', file.replace('dfc', 's2')
        )
        mask_path = os.path.join(self.root, self.directory, 'dfc', file)
        sample = self._load_image(image_path) | self._load_mask(mask_path)

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""2022 IEEE GRSS Data Fusion Contest (DFC2022) dataset."""

import glob
import os
from collections.abc import Callable, Sequence
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import rasterio
import torch
from matplotlib import colors
from matplotlib.figure import Figure
from rasterio.enums import Resampling
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, check_integrity, extract_archive, percentile_normalization


class DFC2022(NonGeoDataset):
    """DFC2022 dataset.

    The `DFC2022 <https://www.grss-ieee.org/community/technical-committees/2022-ieee-grss-data-fusion-contest/>`__
    dataset is used as a benchmark dataset for the 2022 IEEE GRSS Data Fusion Contest
    and extends the MiniFrance dataset for semi-supervised semantic segmentation.
    The dataset consists of a train set containing labeled and unlabeled imagery and an
    unlabeled validation set. The dataset can be downloaded from the
    `IEEEDataPort DFC2022 website <https://ieee-dataport.org/competitions/data-fusion-contest-2022-dfc2022/>`_.

    Dataset features:

    * RGB aerial images at 0.5 m per pixel spatial resolution (~2,000x2,0000 px)
    * DEMs at 1 m per pixel spatial resolution (~1,000x1,0000 px)
    * Masks at 0.5 m per pixel spatial resolution (~2,000x2,0000 px)
    * 16 land use/land cover categories
    * Images collected from the
      `IGN BD ORTHO database <https://geoservices.ign.fr/documentation/donnees/ortho/bdortho/>`_
    * DEMs collected from the
      `IGN RGE ALTI database <https://geoservices.ign.fr/documentation/donnees/alti/rgealti/>`_
    * Labels collected from the
      `UrbanAtlas 2012 database <https://land.copernicus.eu/en/products/urban-atlas/urban-atlas-2012>`_
    * Data collected from 19 regions in France

    Dataset format:

    * images are three-channel geotiffs
    * DEMS are single-channel geotiffs
    * masks are single-channel geotiffs with the pixel values represent the class

    Dataset classes:

    0. No information
    1. Urban fabric
    2. Industrial, commercial, public, military, private and transport units
    3. Mine, dump and construction sites
    4. Artificial non-agricultural vegetated areas
    5. Arable land (annual crops)
    6. Permanent crops
    7. Pastures
    8. Complex and mixed cultivation patterns
    9. Orchards at the fringe of urban classes
    10. Forests
    11. Herbaceous vegetation associations
    12. Open spaces with little or no vegetation
    13. Wetlands
    14. Water
    15. Clouds and Shadows

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1007/s10994-020-05943-y

    .. versionadded:: 0.3
    """

    classes = (
        'No information',
        'Urban fabric',
        'Industrial, commercial, public, military, private and transport units',
        'Mine, dump and construction sites',
        'Artificial non-agricultural vegetated areas',
        'Arable land (annual crops)',
        'Permanent crops',
        'Pastures',
        'Complex and mixed cultivation patterns',
        'Orchards at the fringe of urban classes',
        'Forests',
        'Herbaceous vegetation associations',
        'Open spaces with little or no vegetation',
        'Wetlands',
        'Water',
        'Clouds and Shadows',
    )
    colormap = (
        '#231F20',
        '#DB5F57',
        '#DB9757',
        '#DBD057',
        '#ADDB57',
        '#75DB57',
        '#7BC47B',
        '#58B158',
        '#D4F6D4',
        '#B0E2B0',
        '#008000',
        '#58B0A7',
        '#995D13',
        '#579BDB',
        '#0062FF',
        '#231F20',
    )
    metadata: ClassVar[dict[str, dict[str, str]]] = {
        'train': {
            'filename': 'labeled_train.zip',
            'md5': '2e87d6a218e466dd0566797d7298c7a9',
            'directory': 'labeled_train',
        },
        'train-unlabeled': {
            'filename': 'unlabeled_train.zip',
            'md5': '1016d724bc494b8c50760ae56bb0585e',
            'directory': 'unlabeled_train',
        },
        'val': {
            'filename': 'val.zip',
            'md5': '6ddd9c0f89d8e74b94ea352d4002073f',
            'directory': 'val',
        },
    }

    image_root = 'BDORTHO'
    dem_root = 'RGEALTI'
    target_root = 'UrbanAtlas'

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        checksum: bool = False,
    ) -> None:
        """Initialize a new DFC2022 dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train" or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if ``split`` is invalid
            DatasetNotFoundError: If dataset is not found.
        """
        assert split in self.metadata
        self.root = root
        self.split = split
        self.transforms = transforms
        self.checksum = checksum

        self._verify()

        self.class2idx = {c: i for i, c in enumerate(self.classes)}
        self.files = self._load_files()

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        files = self.files[index]
        image = self._load_image(files['image'])
        dem = self._load_image(files['dem'], shape=image.shape[1:])
        image = torch.cat(tensors=[image, dem], dim=0)

        sample = {'image': image}

        if self.split == 'train':
            mask = self._load_target(files['target'])
            sample['mask'] = mask

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.files)

    def _load_files(self) -> list[dict[str, str]]:
        """Return the paths of the files in the dataset.

        Returns:
            list of dicts containing paths for each pair of image/dem/mask
        """
        directory = os.path.join(self.root, self.metadata[self.split]['directory'])
        images = glob.glob(
            os.path.join(directory, '**', self.image_root, '*.tif'), recursive=True
        )

        files = []
        for image in sorted(images):
            dem = image.replace(self.image_root, self.dem_root)
            dem = f'{os.path.splitext(dem)[0]}_RGEALTI.tif'

            if self.split == 'train':
                target = image.replace(self.image_root, self.target_root)
                target = f'{os.path.splitext(target)[0]}_UA2012.tif'
                files.append(dict(image=image, dem=dem, target=target))
            else:
                files.append(dict(image=image, dem=dem))

        return files

    def _load_image(self, path: Path, shape: Sequence[int] | None = None) -> Tensor:
        """Load a single image.

        Args:
            path: path to the image
            shape: the (h, w) to resample the image to

        Returns:
            the image
        """
        with rasterio.open(path) as f:
            array: np.typing.NDArray[np.float64] = f.read(
                out_shape=shape, out_dtype='float32', resampling=Resampling.bilinear
            )
            tensor = torch.from_numpy(array)
            return tensor

    def _load_target(self, path: Path) -> Tensor:
        """Load the target mask for a single image.

        Args:
            path: path to the image

        Returns:
            the target mask
        """
        with rasterio.open(path) as f:
            array: np.typing.NDArray[np.int_] = f.read(
                indexes=1, out_dtype='int32', resampling=Resampling.bilinear
            )
            tensor = torch.from_numpy(array)
            tensor = tensor.to(torch.long)
            return tensor

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        exists = []
        for split_info in self.metadata.values():
            exists.append(
                os.path.exists(os.path.join(self.root, split_info['directory']))
            )

        if all(exists):
            return

        # Check if .zip files already exists (if so then extract)
        exists = []
        for split_info in self.metadata.values():
            filepath = os.path.join(self.root, split_info['filename'])
            if os.path.isfile(filepath):
                if self.checksum and not check_integrity(filepath, split_info['md5']):
                    raise RuntimeError('Dataset found, but corrupted.')
                exists.append(True)
                extract_archive(filepath)
            else:
                exists.append(False)

        if all(exists):
            return

        raise DatasetNotFoundError(self)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        ncols = 2
        image = sample['image'][:3]
        image = image.to(torch.uint8)
        image_arr = image.permute(1, 2, 0).numpy()

        dem = sample['image'][-1].numpy()
        dem = percentile_normalization(dem, lower=0, upper=100, axis=(0, 1))

        showing_mask = 'mask' in sample
        showing_prediction = 'prediction' in sample

        cmap = colors.ListedColormap(self.colormap)

        if showing_mask:
            mask = sample['mask'].numpy()
            ncols += 1
        if showing_prediction:
            pred = sample['prediction'].numpy()
            ncols += 1

        fig, axs = plt.subplots(nrows=1, ncols=ncols, figsize=(10, ncols * 10))

        axs[0].imshow(image_arr)
        axs[0].axis('off')
        axs[1].imshow(dem)
        axs[1].axis('off')
        if showing_mask:
            axs[2].imshow(mask, cmap=cmap, interpolation='none')
            axs[2].axis('off')
            if showing_prediction:
                axs[3].imshow(pred, cmap=cmap, interpolation='none')
                axs[3].axis('off')
        elif showing_prediction:
            axs[2].imshow(pred, cmap=cmap, interpolation='none')
            axs[2].axis('off')

        if show_titles:
            axs[0].set_title('Image')
            axs[1].set_title('DEM')

            if showing_mask:
                axs[2].set_title('Ground Truth')
                if showing_prediction:
                    axs[3].set_title('Predictions')
            elif showing_prediction:
                axs[2].set_title('Predictions')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""COWC datasets."""

import abc
import csv
import os
from collections.abc import Callable
from typing import cast

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, check_integrity, download_and_extract_archive


class COWC(NonGeoDataset, abc.ABC):
    """Abstract base class for the COWC dataset.

    The `Cars Overhead With Context (COWC) <https://gdo152.llnl.gov/cowc/>`_ data set
    is a large set of annotated cars from overhead. It is useful for training a device
    such as a deep neural network to learn to detect and/or count cars.

    The dataset has the following attributes:

    1. Data from overhead at 15 cm per pixel resolution at ground (all data is EO).
    2. Data from six distinct locations: Toronto, Canada; Selwyn, New Zealand;
       Potsdam and Vaihingen, Germany; Columbus, Ohio and Utah, United States.
    3. 32,716 unique annotated cars. 58,247 unique negative examples.
    4. Intentional selection of hard negative examples.
    5. Established baseline for detection and counting tasks.
    6. Extra testing scenes for use after validation.

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1007/978-3-319-46487-9_48
    """

    @property
    @abc.abstractmethod
    def base_url(self) -> str:
        """Base URL to download dataset from."""

    @property
    @abc.abstractmethod
    def filenames(self) -> tuple[str, ...]:
        """List of files to download."""

    @property
    @abc.abstractmethod
    def md5s(self) -> tuple[str, ...]:
        """List of MD5 checksums of files to download."""

    @property
    @abc.abstractmethod
    def filename(self) -> str:
        """Filename containing train/test split and target labels."""

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new COWC dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train" or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if ``split`` argument is invalid
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in ['train', 'test']

        self.root = root
        self.split = split
        self.transforms = transforms
        self.checksum = checksum

        if download:
            self._download()

        if not self._check_integrity():
            raise DatasetNotFoundError(self)

        self.images = []
        self.targets = []
        with open(
            os.path.join(self.root, self.filename.format(split)),
            encoding='utf-8-sig',
            newline='',
        ) as f:
            reader = csv.reader(f, delimiter=' ')
            for row in reader:
                self.images.append(row[0])
                self.targets.append(row[1])

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        sample = {'image': self._load_image(index), 'label': self._load_target(index)}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.targets)

    def _load_image(self, index: int) -> Tensor:
        """Load a single image.

        Args:
            index: index to return

        Returns:
            the image
        """
        filename = os.path.join(self.root, self.images[index])
        with Image.open(filename) as img:
            array: np.typing.NDArray[np.int_] = np.array(img)
            tensor = torch.from_numpy(array).float()
            # Convert from HxWxC to CxHxW
            tensor = tensor.permute((2, 0, 1))
            return tensor

    def _load_target(self, index: int) -> Tensor:
        """Load a single target.

        Args:
            index: index to return

        Returns:
            the target
        """
        target = int(self.targets[index])
        tensor = torch.tensor(target).float()
        return tensor

    def _check_integrity(self) -> bool:
        """Check integrity of dataset.

        Returns:
            True if dataset files are found and/or MD5s match, else False
        """
        for filename, md5 in zip(self.filenames, self.md5s):
            filepath = os.path.join(self.root, filename)
            if not check_integrity(filepath, md5 if self.checksum else None):
                return False
        return True

    def _download(self) -> None:
        """Download the dataset and extract it."""
        if self._check_integrity():
            print('Files already downloaded and verified')
            return

        for filename, md5 in zip(self.filenames, self.md5s):
            download_and_extract_archive(
                self.base_url + filename,
                self.root,
                filename=filename,
                md5=md5 if self.checksum else None,
            )

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        .. versionadded:: 0.2
        """
        image = sample['image']
        label = cast(str, sample['label'].item())

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = cast(str, sample['prediction'].item())
        else:
            prediction = None

        fig, ax = plt.subplots(1, 1, figsize=(10, 10))
        ax.imshow(image.permute(1, 2, 0))
        ax.axis('off')

        if show_titles:
            title = f'Label: {label}'
            if prediction is not None:
                title += f'\nPrediction: {prediction}'
            ax.set_title(title)

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig


class COWCCounting(COWC):
    """COWC Dataset for car counting."""

    base_url = (
        'https://gdo152.llnl.gov/cowc/download/cowc/datasets/patch_sets/counting/'
    )
    filenames = (
        'COWC_train_list_64_class.txt.bz2',
        'COWC_test_list_64_class.txt.bz2',
        'COWC_Counting_Toronto_ISPRS.tbz',
        'COWC_Counting_Selwyn_LINZ.tbz',
        'COWC_Counting_Potsdam_ISPRS.tbz',
        'COWC_Counting_Vaihingen_ISPRS.tbz',
        'COWC_Counting_Columbus_CSUAV_AFRL.tbz',
        'COWC_Counting_Utah_AGRC.tbz',
    )
    md5s = (
        '187543d20fa6d591b8da51136e8ef8fb',
        '930cfd6e160a7b36db03146282178807',
        'bc2613196dfa93e66d324ae43e7c1fdb',
        'ea842ae055f5c74d0d933d2194764545',
        '19a77ab9932b722ef52b197d70e68ce7',
        '4009c1e420566390746f5b4db02afdb9',
        'daf8033c4e8ceebbf2c3cac3fabb8b10',
        '777ec107ed2a3d54597a739ce74f95ad',
    )
    filename = 'COWC_{}_list_64_class.txt'


class COWCDetection(COWC):
    """COWC Dataset for car detection."""

    base_url = (
        'https://gdo152.llnl.gov/cowc/download/cowc/datasets/patch_sets/detection/'
    )
    filenames = (
        'COWC_train_list_detection.txt.bz2',
        'COWC_test_list_detection.txt.bz2',
        'COWC_Detection_Toronto_ISPRS.tbz',
        'COWC_Detection_Selwyn_LINZ.tbz',
        'COWC_Detection_Potsdam_ISPRS.tbz',
        'COWC_Detection_Vaihingen_ISPRS.tbz',
        'COWC_Detection_Columbus_CSUAV_AFRL.tbz',
        'COWC_Detection_Utah_AGRC.tbz',
    )
    md5s = (
        'c954a5a3dac08c220b10cfbeec83893c',
        'c6c2d0a78f12a2ad88b286b724a57c1a',
        '11af24f43b198b0f13c8e94814008a48',
        '22fd37a86961010f5d519a7da0e1fc72',
        'bf053545cc1915d8b6597415b746fe48',
        '23945d5b22455450a938382ccc2a8b27',
        'f40522dc97bea41b10117d4a5b946a6f',
        '195da7c9443a939a468c9f232fd86ee3',
    )
    filename = 'COWC_{}_list_detection.txt'


# TODO: add COCW-M datasets:
#
# * https://gdo152.llnl.gov/cowc/download/cowc-m/datasets/
# * https://github.com/LLNL/cowc
#
# Same as COCW datasets, but instead of binary classification there are 4 car classes:
#
# 1. Sedan
# 2. Pickup
# 3. Other
# 4. Unknown
#
# May need new abstract base class. Will need subclasses for different patch sizes.



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""QuakeSet dataset."""

import os
from collections.abc import Callable
from typing import Any, ClassVar, cast

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_url, lazy_import, percentile_normalization


class QuakeSet(NonGeoDataset):
    """QuakeSet dataset.

    `QuakeSet <https://huggingface.co/datasets/DarthReca/quakeset>`__
    is a dataset for Earthquake Change Detection and Magnitude Estimation and is used
    for the Seismic Monitoring and Analysis (SMAC) ECML-PKDD 2024 Discovery Challenge.

    Dataset features:

    * Sentinel-1 SAR imagery
    * before/pre/post imagery of areas affected by earthquakes
    * 2 SAR bands (VV/VH)
    * 3,327 pairs of pre and post images with 5 m per pixel resolution (512x512 px)
    * 2 classification labels (unaffected / affected by earthquake)
    * pre/post image pairs represent earthquake affected areas
    * before/pre image pairs represent hard negative unaffected areas
    * earthquake magnitudes for each sample

    Dataset format:

    * single hdf5 dataset containing images, magnitudes, hypercenters, and splits

    Dataset classes:

    0. unaffected area
    1. earthquake affected area

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2403.18116

    .. note::

       This dataset requires the following additional library to be installed:

       * `h5py <https://pypi.org/project/h5py/>`_ to load the dataset

    .. versionadded:: 0.6
    """

    filename = 'earthquakes.h5'
    url = 'https://hf.co/datasets/DarthReca/quakeset/resolve/bead1d25fb9979dbf703f9ede3e8b349f73b29f7/earthquakes.h5'
    md5 = '76fc7c76b7ca56f4844d852e175e1560'
    splits: ClassVar[dict[str, str]] = {
        'train': 'train',
        'val': 'validation',
        'test': 'test',
    }
    classes = ('unaffected_area', 'earthquake_affected_area')

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new QuakeSet dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train", "val", or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: If ``split`` argument is invalid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
            DependencyNotFoundError: If h5py is not installed.
        """
        lazy_import('h5py')

        assert split in self.splits

        self.root = root
        self.split = split
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self.filepath = os.path.join(root, self.filename)
        self._verify()
        self.data = self._load_data()

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            sample containing image and mask
        """
        image = self._load_image(index)
        label = torch.tensor(self.data[index]['label'])
        magnitude = torch.tensor(self.data[index]['magnitude'])

        sample = {'image': image, 'label': label, 'magnitude': magnitude}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.data)

    def _load_data(self) -> list[dict[str, Any]]:
        """Return the metadata for a given split.

        Returns:
            the sample keys, patches, images, labels, and magnitudes
        """
        h5py = lazy_import('h5py')
        data = []
        with h5py.File(self.filepath) as f:
            for k in sorted(f.keys()):
                if f[k].attrs['split'] != self.splits[self.split]:
                    continue

                for patch in sorted(f[k].keys()):
                    if patch not in ['x', 'y']:
                        # positive sample
                        magnitude = float(f[k].attrs['magnitude'])
                        data.append(
                            dict(
                                key=k,
                                patch=patch,
                                images=('pre', 'post'),
                                label=1,
                                magnitude=magnitude,
                            )
                        )

                        # hard negative sample
                        if 'before' in f[k][patch].keys():
                            data.append(
                                dict(
                                    key=k,
                                    patch=patch,
                                    images=('before', 'pre'),
                                    label=0,
                                    magnitude=0.0,
                                )
                            )
        return data

    def _load_image(self, index: int) -> Tensor:
        """Load a single image.

        Args:
            index: index to return

        Returns:
            the image
        """
        h5py = lazy_import('h5py')

        key = self.data[index]['key']
        patch = self.data[index]['patch']
        images = self.data[index]['images']

        with h5py.File(self.filepath) as f:
            pre_array = f[key][patch][images[0]][:]
            pre_array = np.nan_to_num(pre_array, nan=0)
            post_array = f[key][patch][images[1]][:]
            post_array = np.nan_to_num(post_array, nan=0)
            array = np.concatenate([pre_array, post_array], axis=-1)
            array = array.astype(np.float32)

        tensor = torch.from_numpy(array)
        # Convert from HxWxC to CxHxW
        tensor = tensor.permute((2, 0, 1))
        return tensor

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        if os.path.exists(self.filepath):
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        if not os.path.exists(self.filepath):
            download_url(
                self.url,
                self.root,
                filename=self.filename,
                md5=self.md5 if self.checksum else None,
            )

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        image = sample['image'].permute((1, 2, 0)).numpy()
        label = cast(int, sample['label'].item())
        label_class = self.classes[label]

        # Create false color image for image1
        vv = percentile_normalization(image[..., 0]) + 1e-16
        vh = percentile_normalization(image[..., 1]) + 1e-16
        fci1 = np.stack([vv, vh, vv / vh], axis=-1).clip(0, 1)

        # Create false color image for image2
        vv = percentile_normalization(image[..., 2]) + 1e-16
        vh = percentile_normalization(image[..., 3]) + 1e-16
        fci2 = np.stack([vv, vh, vv / vh], axis=-1).clip(0, 1)

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = cast(int, sample['prediction'].item())
            prediction_class = self.classes[prediction]

        ncols = 2
        fig, axs = plt.subplots(
            nrows=1, ncols=ncols, figsize=(ncols * 5, 10), sharex=True
        )

        axs[0].imshow(fci1)
        axs[0].axis('off')
        axs[0].set_title('Image Pre')
        axs[1].imshow(fci2)
        axs[1].axis('off')
        axs[1].set_title('Image Post')

        if show_titles:
            title = f'Label: {label_class}'
            if 'magnitude' in sample:
                magnitude = cast(float, sample['magnitude'].item())
                title += f' | Magnitude: {magnitude:.2f}'
            if showing_predictions:
                title += f'\nPrediction: {prediction_class}'
            fig.supxlabel(title, y=0.22)

        if suptitle is not None:
            fig.suptitle(suptitle, y=0.8)

        fig.tight_layout()

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""TorchGeo samplers."""

import abc
from collections.abc import Callable, Iterable, Iterator
from functools import partial

import torch
from rtree.index import Index, Property
from torch import Generator
from torch.utils.data import Sampler

from ..datasets import BoundingBox, GeoDataset
from .constants import Units
from .utils import _to_tuple, get_random_bounding_box, tile_to_chips


class GeoSampler(Sampler[BoundingBox], abc.ABC):
    """Abstract base class for sampling from :class:`~torchgeo.datasets.GeoDataset`.

    Unlike PyTorch's :class:`~torch.utils.data.Sampler`, :class:`GeoSampler`
    returns enough geospatial information to uniquely index any
    :class:`~torchgeo.datasets.GeoDataset`. This includes things like latitude,
    longitude, height, width, projection, coordinate system, and time.
    """

    def __init__(self, dataset: GeoDataset, roi: BoundingBox | None = None) -> None:
        """Initialize a new Sampler instance.

        Args:
            dataset: dataset to index from
            roi: region of interest to sample from (minx, maxx, miny, maxy, mint, maxt)
                (defaults to the bounds of ``dataset.index``)
        """
        if roi is None:
            self.index = dataset.index
            roi = BoundingBox(*self.index.bounds)
        else:
            self.index = Index(interleaved=False, properties=Property(dimension=3))
            hits = dataset.index.intersection(tuple(roi), objects=True)
            for hit in hits:
                bbox = BoundingBox(*hit.bounds) & roi
                self.index.insert(hit.id, tuple(bbox), hit.object)

        self.res = dataset.res
        self.roi = roi

    @abc.abstractmethod
    def __iter__(self) -> Iterator[BoundingBox]:
        """Return the index of a dataset.

        Returns:
            (minx, maxx, miny, maxy, mint, maxt) coordinates to index a dataset
        """


class RandomGeoSampler(GeoSampler):
    """Samples elements from a region of interest randomly.

    This is particularly useful during training when you want to maximize the size of
    the dataset and return as many random :term:`chips <chip>` as possible. Note that
    randomly sampled chips may overlap.

    This sampler is not recommended for use with tile-based datasets. Use
    :class:`RandomBatchGeoSampler` instead.
    """

    def __init__(
        self,
        dataset: GeoDataset,
        size: tuple[float, float] | float,
        length: int | None = None,
        roi: BoundingBox | None = None,
        units: Units = Units.PIXELS,
        generator: Generator | None = None,
    ) -> None:
        """Initialize a new Sampler instance.

        The ``size`` argument can either be:

        * a single ``float`` - in which case the same value is used for the height and
          width dimension
        * a ``tuple`` of two floats - in which case, the first *float* is used for the
          height dimension, and the second *float* for the width dimension

        .. versionchanged:: 0.3
           Added ``units`` parameter, changed default to pixel units

        .. versionchanged:: 0.4
           ``length`` parameter is now optional, a reasonable default will be used

        .. versionadded:: 0.7
            The *generator* parameter.

        Args:
            dataset: dataset to index from
            size: dimensions of each :term:`patch`
            length: number of random samples to draw per epoch
                (defaults to approximately the maximal number of non-overlapping
                :term:`chips <chip>` of size ``size`` that could be sampled from
                the dataset)
            roi: region of interest to sample from (minx, maxx, miny, maxy, mint, maxt)
                (defaults to the bounds of ``dataset.index``)
            units: defines if ``size`` is in pixel or CRS units
            generator: pseudo-random number generator (PRNG).
        """
        super().__init__(dataset, roi)
        self.size = _to_tuple(size)

        if units == Units.PIXELS:
            self.size = (self.size[0] * self.res[1], self.size[1] * self.res[0])

        self.generator = generator
        self.length = 0
        self.hits = []
        areas = []
        for hit in self.index.intersection(tuple(self.roi), objects=True):
            bounds = BoundingBox(*hit.bounds)
            if (
                bounds.maxx - bounds.minx >= self.size[1]
                and bounds.maxy - bounds.miny >= self.size[0]
            ):
                if bounds.area > 0:
                    rows, cols = tile_to_chips(bounds, self.size)
                    self.length += rows * cols
                else:
                    self.length += 1
                self.hits.append(hit)
                areas.append(bounds.area)
        if length is not None:
            self.length = length

        # torch.multinomial requires float probabilities > 0
        self.areas = torch.tensor(areas, dtype=torch.float)
        if torch.sum(self.areas) == 0:
            self.areas += 1

    def __iter__(self) -> Iterator[BoundingBox]:
        """Return the index of a dataset.

        Returns:
            (minx, maxx, miny, maxy, mint, maxt) coordinates to index a dataset
        """
        for _ in range(len(self)):
            # Choose a random tile, weighted by area
            idx = torch.multinomial(self.areas, 1)
            hit = self.hits[idx]
            bounds = BoundingBox(*hit.bounds)

            # Choose a random index within that tile
            bounding_box = get_random_bounding_box(
                bounds, self.size, self.res, self.generator
            )

            yield bounding_box

    def __len__(self) -> int:
        """Return the number of samples in a single epoch.

        Returns:
            length of the epoch
        """
        return self.length


class GridGeoSampler(GeoSampler):
    """Samples elements in a grid-like fashion.

    This is particularly useful during evaluation when you want to make predictions for
    an entire region of interest. You want to minimize the amount of redundant
    computation by minimizing overlap between :term:`chips <chip>`.

    Usually the stride should be slightly smaller than the chip size such that each chip
    has some small overlap with surrounding chips. This is used to prevent `stitching
    artifacts <https://arxiv.org/abs/1805.12219>`_ when combining each prediction patch.
    The overlap between each chip (``chip_size - stride``) should be approximately equal
    to the `receptive field <https://distill.pub/2019/computing-receptive-fields/>`_ of
    the CNN.
    """

    def __init__(
        self,
        dataset: GeoDataset,
        size: tuple[float, float] | float,
        stride: tuple[float, float] | float,
        roi: BoundingBox | None = None,
        units: Units = Units.PIXELS,
    ) -> None:
        """Initialize a new Sampler instance.

        The ``size`` and ``stride`` arguments can either be:

        * a single ``float`` - in which case the same value is used for the height and
          width dimension
        * a ``tuple`` of two floats - in which case, the first *float* is used for the
          height dimension, and the second *float* for the width dimension

        .. versionchanged:: 0.3
           Added ``units`` parameter, changed default to pixel units

        Args:
            dataset: dataset to index from
            size: dimensions of each :term:`patch`
            stride: distance to skip between each patch
            roi: region of interest to sample from (minx, maxx, miny, maxy, mint, maxt)
                (defaults to the bounds of ``dataset.index``)
            units: defines if ``size`` and ``stride`` are in pixel or CRS units
        """
        super().__init__(dataset, roi)
        self.size = _to_tuple(size)
        self.stride = _to_tuple(stride)

        if units == Units.PIXELS:
            self.size = (self.size[0] * self.res[1], self.size[1] * self.res[0])
            self.stride = (self.stride[0] * self.res[1], self.stride[1] * self.res[0])

        self.hits = []
        for hit in self.index.intersection(tuple(self.roi), objects=True):
            bounds = BoundingBox(*hit.bounds)
            if (
                bounds.maxx - bounds.minx >= self.size[1]
                and bounds.maxy - bounds.miny >= self.size[0]
            ):
                self.hits.append(hit)

        self.length = 0
        for hit in self.hits:
            bounds = BoundingBox(*hit.bounds)
            rows, cols = tile_to_chips(bounds, self.size, self.stride)
            self.length += rows * cols

    def __iter__(self) -> Iterator[BoundingBox]:
        """Return the index of a dataset.

        Returns:
            (minx, maxx, miny, maxy, mint, maxt) coordinates to index a dataset
        """
        # For each tile...
        for hit in self.hits:
            bounds = BoundingBox(*hit.bounds)
            rows, cols = tile_to_chips(bounds, self.size, self.stride)
            mint = bounds.mint
            maxt = bounds.maxt

            # For each row...
            for i in range(rows):
                miny = bounds.miny + i * self.stride[0]
                maxy = miny + self.size[0]

                # For each column...
                for j in range(cols):
                    minx = bounds.minx + j * self.stride[1]
                    maxx = minx + self.size[1]

                    yield BoundingBox(minx, maxx, miny, maxy, mint, maxt)

    def __len__(self) -> int:
        """Return the number of samples over the ROI.

        Returns:
            number of patches that will be sampled
        """
        return self.length


class PreChippedGeoSampler(GeoSampler):
    """Samples entire files at a time.

    This is particularly useful for datasets that contain geospatial metadata
    and subclass :class:`~torchgeo.datasets.GeoDataset` but have already been
    pre-processed into :term:`chips <chip>`.

    This sampler should not be used with :class:`~torchgeo.datasets.NonGeoDataset`.
    You may encounter problems when using an :term:`ROI <region of interest (ROI)>`
    that partially intersects with one of the file bounding boxes, when using an
    :class:`~torchgeo.datasets.IntersectionDataset`, or when each file is in a
    different CRS. These issues can be solved by adding padding.
    """

    def __init__(
        self,
        dataset: GeoDataset,
        roi: BoundingBox | None = None,
        shuffle: bool = False,
        generator: torch.Generator | None = None,
    ) -> None:
        """Initialize a new Sampler instance.

        .. versionadded:: 0.3

        .. versionadded:: 0.7
            The *generator* parameter.

        Args:
            dataset: dataset to index from
            roi: region of interest to sample from (minx, maxx, miny, maxy, mint, maxt)
                (defaults to the bounds of ``dataset.index``)
            shuffle: if True, reshuffle data at every epoch
            generator: pseudo-random number generator (PRNG) used in combination with shuffle.

        """
        super().__init__(dataset, roi)
        self.shuffle = shuffle
        self.generator = generator

        self.hits = []
        for hit in self.index.intersection(tuple(self.roi), objects=True):
            self.hits.append(hit)

    def __iter__(self) -> Iterator[BoundingBox]:
        """Return the index of a dataset.

        Returns:
            (minx, maxx, miny, maxy, mint, maxt) coordinates to index a dataset
        """
        generator: Callable[[int], Iterable[int]] = range
        if self.shuffle:
            generator = partial(torch.randperm, generator=self.generator)

        for idx in generator(len(self)):
            yield BoundingBox(*self.hits[idx].bounds)

    def __len__(self) -> int:
        """Return the number of samples over the ROI.

        Returns:
            number of patches that will be sampled
        """
        return len(self.hits)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""National Agriculture Imagery Program (NAIP) dataset."""

from typing import Any

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from .geo import RasterDataset


class NAIP(RasterDataset):
    """National Agriculture Imagery Program (NAIP) dataset.

    The `National Agriculture Imagery Program (NAIP)
    <https://catalog.data.gov/dataset/national-agriculture-imagery-program-naip>`_
    acquires aerial imagery during the agricultural growing seasons in the continental
    U.S. A primary goal of the NAIP program is to make digital ortho photography
    available to governmental agencies and the public within a year of acquisition.

    NAIP is administered by the USDA's Farm Service Agency (FSA) through the Aerial
    Photography Field Office in Salt Lake City. This "leaf-on" imagery is used as a base
    layer for GIS programs in FSA's County Service Centers, and is used to maintain the
    Common Land Unit (CLU) boundaries.

    If you use this dataset in your research, please cite it using the following format:

    * https://www.fisheries.noaa.gov/inport/item/49508/citation
    """

    # https://www.nrcs.usda.gov/Internet/FSE_DOCUMENTS/nrcs141p2_015644.pdf
    # https://planetarycomputer.microsoft.com/dataset/naip#Storage-Documentation
    filename_glob = 'm_*.*'
    filename_regex = r"""
        ^m
        _(?P<quadrangle>\d+)
        _(?P<quarter_quad>[a-z]+)
        _(?P<utm_zone>\d+)
        _(?P<resolution>\d+)
        _(?P<date>\d+)
        (?:_(?P<processing_date>\d+))?
        \..*$
    """

    # Plotting
    all_bands = ('R', 'G', 'B', 'NIR')
    rgb_bands = ('R', 'G', 'B')

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`RasterDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        .. versionchanged:: 0.3
           Method now takes a sample dict, not a Tensor. Additionally, possible to
           show subplot titles and/or use a custom suptitle.
        """
        image = sample['image'][0:3, :, :].permute(1, 2, 0)

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(4, 4))

        ax.imshow(image)
        ax.axis('off')
        if show_titles:
            ax.set_title('Image')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Foreground-Aware Relation Network (FarSeg) implementations."""

import math
from collections import OrderedDict
from typing import cast

import torch.nn.functional as F
import torchvision
from torch import Tensor
from torch.nn.modules import (
    BatchNorm2d,
    Conv2d,
    Identity,
    Module,
    ModuleList,
    ReLU,
    Sequential,
    Sigmoid,
    UpsamplingBilinear2d,
)
from torchvision.models import resnet
from torchvision.ops import FeaturePyramidNetwork as FPN


class FarSeg(Module):
    """Foreground-Aware Relation Network (FarSeg).

    This model can be used for binary- or multi-class object segmentation, such as
    building, road, ship, and airplane segmentation. It can be also extended as a change
    detection model. It features a foreground-scene relation module to model the
    relation between scene embedding, object context, and object feature, thus improving
    the discrimination of object feature representation.

    If you use this model in your research, please cite the following paper:

    * https://arxiv.org/pdf/2011.09766
    """

    def __init__(
        self,
        backbone: str = 'resnet50',
        classes: int = 16,
        backbone_pretrained: bool = True,
    ) -> None:
        """Initialize a new FarSeg model.

        Args:
            backbone: name of ResNet backbone, one of ["resnet18", "resnet34",
                "resnet50", "resnet101"]
            classes: number of output segmentation classes
            backbone_pretrained: whether to use pretrained weight for backbone
        """
        super().__init__()
        if backbone in ['resnet18', 'resnet34']:
            max_channels = 512
        elif backbone in ['resnet50', 'resnet101']:
            max_channels = 2048
        else:
            raise ValueError(f'unknown backbone: {backbone}.')
        kwargs = {}
        if backbone_pretrained:
            kwargs = {
                'weights': getattr(
                    torchvision.models, f'ResNet{backbone[6:]}_Weights'
                ).DEFAULT
            }
        else:
            kwargs = {'weights': None}

        self.backbone = getattr(resnet, backbone)(**kwargs)

        self.fpn = FPN(
            in_channels_list=[max_channels // (2 ** (3 - i)) for i in range(4)],
            out_channels=256,
        )
        self.fsr = _FSRelation(max_channels, [256] * 4, 256)
        self.decoder = _LightWeightDecoder(256, 128, classes)

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass of the model.

        Args:
            x: input image

        Returns:
            output prediction
        """
        x = self.backbone.conv1(x)
        x = self.backbone.bn1(x)
        x = self.backbone.relu(x)
        x = self.backbone.maxpool(x)

        c2 = self.backbone.layer1(x)
        c3 = self.backbone.layer2(c2)
        c4 = self.backbone.layer3(c3)
        c5 = self.backbone.layer4(c4)
        features = [c2, c3, c4, c5]

        coarsest_features = features[-1]
        scene_embedding = F.adaptive_avg_pool2d(coarsest_features, 1)
        fpn_features = self.fpn(
            OrderedDict({f'c{i + 2}': features[i] for i in range(4)})
        )
        features = [v for k, v in fpn_features.items()]
        features = self.fsr(scene_embedding, features)

        logit = self.decoder(features)

        return cast(Tensor, logit)


class _FSRelation(Module):
    """F-S Relation module."""

    def __init__(
        self,
        scene_embedding_channels: int,
        in_channels_list: list[int],
        out_channels: int,
    ) -> None:
        """Initialize the _FSRelation module.

        Args:
            scene_embedding_channels: number of scene embedding channels
            in_channels_list: a list of input channels
            out_channels: number of output channels
        """
        super().__init__()

        self.scene_encoder = ModuleList(
            [
                Sequential(
                    Conv2d(scene_embedding_channels, out_channels, 1),
                    ReLU(True),
                    Conv2d(out_channels, out_channels, 1),
                )
                for _ in range(len(in_channels_list))
            ]
        )

        self.content_encoders = ModuleList()
        self.feature_reencoders = ModuleList()
        for c in in_channels_list:
            self.content_encoders.append(
                Sequential(
                    Conv2d(c, out_channels, 1), BatchNorm2d(out_channels), ReLU(True)
                )
            )
            self.feature_reencoders.append(
                Sequential(
                    Conv2d(c, out_channels, 1), BatchNorm2d(out_channels), ReLU(True)
                )
            )

        self.normalizer = Sigmoid()

    def forward(self, scene_feature: Tensor, features: list[Tensor]) -> list[Tensor]:
        """Forward pass of the model."""
        # [N, C, H, W]
        content_feats = [
            c_en(p_feat) for c_en, p_feat in zip(self.content_encoders, features)
        ]
        scene_feats = [op(scene_feature) for op in self.scene_encoder]
        relations = [
            self.normalizer((sf * cf).sum(dim=1, keepdim=True))
            for sf, cf in zip(scene_feats, content_feats)
        ]

        p_feats = [op(p_feat) for op, p_feat in zip(self.feature_reencoders, features)]

        refined_feats = [r * p for r, p in zip(relations, p_feats)]

        return refined_feats


class _LightWeightDecoder(Module):
    """Light Weight Decoder."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_classes: int,
        in_feature_output_strides: list[int] = [4, 8, 16, 32],
        out_feature_output_stride: int = 4,
    ) -> None:
        """Initialize the _LightWeightDecoder module.

        Args:
            in_channels: number of channels of input feature maps
            out_channels: number of channels of output feature maps
            num_classes: number of output segmentation classes
            in_feature_output_strides: output stride of input feature maps at different
                levels
            out_feature_output_stride: output stride of output feature maps
        """
        super().__init__()

        self.blocks = ModuleList()
        for in_feat_os in in_feature_output_strides:
            num_upsample = int(math.log2(int(in_feat_os))) - int(
                math.log2(int(out_feature_output_stride))
            )
            num_layers = num_upsample if num_upsample != 0 else 1
            self.blocks.append(
                Sequential(
                    *[
                        Sequential(
                            Conv2d(
                                in_channels if idx == 0 else out_channels,
                                out_channels,
                                3,
                                1,
                                1,
                                bias=False,
                            ),
                            BatchNorm2d(out_channels),
                            ReLU(inplace=True),
                            (
                                UpsamplingBilinear2d(scale_factor=2)
                                if num_upsample != 0
                                else Identity()
                            ),
                        )
                        for idx in range(num_layers)
                    ]
                )
            )

        self.classifier = Sequential(
            Conv2d(out_channels, num_classes, 3, 1, 1),
            UpsamplingBilinear2d(scale_factor=4),
        )

    def forward(self, features: list[Tensor]) -> Tensor:
        """Forward pass of the model."""
        inner_feat_list = []
        for idx, block in enumerate(self.blocks):
            decoder_feat = block(features[idx])
            inner_feat_list.append(decoder_feat)

        out_feat = sum(inner_feat_list) / len(inner_feat_list)
        out_feat = self.classifier(out_feat)

        return cast(Tensor, out_feat)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Substation datamodule."""

from typing import Any

import kornia.augmentation as K
import torch
from torch.utils.data import random_split

from ..datasets import Substation
from .geo import NonGeoDataModule


class SubstationDataModule(NonGeoDataModule):
    """Substation Data Module with train-test split and transformations.

    .. versionadded:: 0.7
    """

    def __init__(
        self,
        batch_size: int = 64,
        num_workers: int = 0,
        val_split_pct: float = 0.2,
        test_split_pct: float = 0.2,
        size: int = 256,
        **kwargs: Any,
    ) -> None:
        """Initialize a new SubstationDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for data loading.
            val_split_pct: Percentage of data to use for validation.
            test_split_pct: Percentage of data to use for testing.
            size: Size of the input images.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.Substation`.
        """
        super().__init__(Substation, batch_size, num_workers, **kwargs)
        self.val_split_pct = val_split_pct
        self.test_split_pct = test_split_pct
        self.train_aug = K.AugmentationSequential(
            K.Resize(size),
            K.RandomHorizontalFlip(p=0.5),
            K.RandomVerticalFlip(p=0.5),
            data_keys=None,
            keepdim=True,
        )
        self.aug = K.AugmentationSequential(
            K.Resize(size), data_keys=None, keepdim=True
        )

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: One of 'fit', 'validate', 'test', or 'predict'.
        """
        dataset = Substation(**self.kwargs)

        generator = torch.Generator().manual_seed(0)
        total_len = len(dataset)
        val_len = int(total_len * self.val_split_pct)
        test_len = int(total_len * self.test_split_pct)
        train_len = total_len - val_len - test_len

        self.train_dataset, self.val_dataset, self.test_dataset = random_split(
            dataset, [train_len, val_len, test_len], generator
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Loss functions for learing on the prior."""

import torch
import torch.nn.functional as F
from torch.nn.modules import Module


class QRLoss(Module):
    """The QR (forward) loss between class probabilities and predictions.

    This loss is defined in `'Resolving label uncertainty with implicit generative
    models' <https://openreview.net/forum?id=AEa_UepnMDX>`_.

    .. versionadded:: 0.2
    """

    def forward(self, probs: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Computes the QR (forwards) loss on prior.

        Args:
            probs: probabilities of predictions, expected shape B x C x H x W.
            target: prior probabilities, expected shape B x C x H x W.

        Returns:
            qr loss
        """
        q = probs
        q_bar = q.mean(dim=(0, 2, 3))
        qbar_log_S = (q_bar * torch.log(q_bar)).sum()

        q_log_p = torch.einsum('bcxy,bcxy->bxy', q, torch.log(target)).mean()

        loss = qbar_log_S - q_log_p
        return loss


class RQLoss(Module):
    """The RQ (backwards) loss between class probabilities and predictions.

    This loss is defined in `'Resolving label uncertainty with implicit generative
    models' <https://openreview.net/forum?id=AEa_UepnMDX>`_.

    .. versionadded:: 0.2
    """

    def forward(self, probs: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Computes the RQ (backwards) loss on prior.

        Args:
            probs: probabilities of predictions, expected shape B x C x H x W
            target: prior probabilities, expected shape B x C x H x W

        Returns:
            qr loss
        """
        q = probs

        # manually normalize due to https://github.com/pytorch/pytorch/issues/70100
        z = q / q.norm(p=1, dim=(0, 2, 3), keepdim=True).clamp_min(1e-12).expand_as(q)
        r = F.normalize(z * target, p=1, dim=1)

        loss = torch.einsum('bcxy,bcxy->bxy', r, torch.log(r) - torch.log(q)).mean()

        return loss



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""BioMassters Dataset."""

import os
from collections.abc import Sequence

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, percentile_normalization


class BioMassters(NonGeoDataset):
    """BioMassters Dataset for Aboveground Biomass prediction.

    Dataset intended for Aboveground Biomass (AGB) prediction
    over Finnish forests based on Sentinel 1 and 2 data with
    corresponding target AGB mask values generated by Light Detection
    and Ranging (LiDAR).

    Dataset Format:

    * .tif files for Sentinel 1 and 2 data
    * .tif file for pixel wise AGB target mask
    * .csv files for metadata regarding features and targets

    Dataset Features:

    * 13,000 target AGB masks of size (256x256px)
    * 12 months of data per target mask
    * Sentinel 1 and Sentinel 2 data for each location
    * Sentinel 1 available for every month
    * Sentinel 2 available for almost every month
      (not available for every month due to ESA acquisition halt over the region
      during particular periods)

    If you use this dataset in your research, please cite the following paper:

    * https://nascetti-a.github.io/BioMasster/

    .. note::
        This dataset can be downloaded from `Torchgeo Hugging Face Hub <https://hf.co/datasets/torchgeo/biomassters>`_.

    .. versionadded:: 0.5
    """

    valid_splits = ('train', 'test')
    valid_sensors = ('S1', 'S2')

    metadata_filename = 'biomassters_features_metadata.csv'

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        sensors: Sequence[str] = ['S1', 'S2'],
        as_time_series: bool = False,
    ) -> None:
        """Initialize a new instance of BioMassters dataset.

        If ``as_time_series=False`` (the default), each time step becomes its own
        sample with the target being shared across multiple samples.

        Args:
            root: root directory where dataset can be found
            split: train or test split
            sensors: which sensors to consider for the sample, Sentinel 1 and/or
                Sentinel 2 ('S1', 'S2')
            as_time_series: whether or not to return all available
                time-steps or just a single one for a given target location

        Raises:
            AssertionError: if ``split`` or ``sensors`` is invalid
            DatasetNotFoundError: If dataset is not found.
        """
        self.root = root

        assert split in self.valid_splits, (
            f'Please choose one of the valid splits: {self.valid_splits}.'
        )
        self.split = split

        assert set(sensors).issubset(set(self.valid_sensors)), (
            f'Please choose a subset of valid sensors: {self.valid_sensors}.'
        )
        self.sensors = sensors
        self.as_time_series = as_time_series

        self._verify()

        # open metadata csv files
        self.df = pd.read_csv(os.path.join(self.root, self.metadata_filename))

        # filter sensors
        self.df = self.df[self.df['satellite'].isin(self.sensors)]

        # filter split
        self.df = self.df[self.df['split'] == self.split]

        # generate numerical month from filename since first month is September
        # and has numerical index of 0
        self.df['num_month'] = (
            self.df['filename']
            .str.split('_', expand=True)[2]
            .str.split('.', expand=True)[0]
            .astype(int)
        )

        # set dataframe index depending on the task for easier indexing
        if self.as_time_series:
            self.df['num_index'] = self.df.groupby(['chip_id']).ngroup()
        else:
            filter_df = (
                self.df.groupby(['chip_id', 'month'])['satellite'].count().reset_index()
            )
            filter_df = filter_df[filter_df['satellite'] == len(self.sensors)].drop(
                'satellite', axis=1
            )
            # guarantee that each sample has corresponding number of images available
            self.df = self.df.merge(filter_df, on=['chip_id', 'month'], how='inner')

            self.df['num_index'] = self.df.groupby(['chip_id', 'month']).ngroup()

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and labels at that index

        Raises:
            IndexError: if index is out of range of the dataset
        """
        sample_df = self.df[self.df['num_index'] == index].copy()

        # sort by satellite and month to return correct order
        sample_df.sort_values(
            by=['satellite', 'num_month'], inplace=True, ascending=True
        )

        filepaths = sample_df['filename'].tolist()
        sample: dict[str, Tensor] = {}
        for sens in self.sensors:
            sens_filepaths = [fp for fp in filepaths if sens in fp]
            sample[f'image_{sens}'] = self._load_input(sens_filepaths)

        if self.split == 'train':
            sample['label'] = self._load_target(
                sample_df['corresponding_agbm'].unique()[0]
            )

        return sample

    def __len__(self) -> int:
        """Return the length of the dataset.

        Returns:
            length of the dataset
        """
        return len(self.df['num_index'].unique())

    def _load_input(self, filenames: list[Path]) -> Tensor:
        """Load the input imagery at the index.

        Args:
            filenames: list of filenames corresponding to input

        Returns:
            input image
        """
        filepaths = [
            os.path.join(self.root, f'{self.split}_features', f) for f in filenames
        ]
        arr_list = [rasterio.open(fp).read() for fp in filepaths]
        if self.as_time_series:
            arr = np.stack(arr_list, axis=0)
        else:
            arr = np.concatenate(arr_list, axis=0)
        return torch.tensor(arr.astype(np.int32))

    def _load_target(self, filename: Path) -> Tensor:
        """Load the target mask at the index.

        Args:
            filename: filename of target to index

        Returns:
            target mask
        """
        with rasterio.open(os.path.join(self.root, 'train_agbm', filename), 'r') as src:
            arr: np.typing.NDArray[np.float64] = src.read()

        target = torch.from_numpy(arr).float()
        return target

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        exists = []

        filenames = [f'{self.split}_features', self.metadata_filename]
        for filename in filenames:
            pathname = os.path.join(self.root, filename)
            exists.append(os.path.exists(pathname))
        if all(exists):
            return

        raise DatasetNotFoundError(self)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample return by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        ncols = len(self.sensors) + 1

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            ncols += 1

        fig, axs = plt.subplots(1, ncols=ncols, figsize=(5 * ncols, 10))
        for idx, sens in enumerate(self.sensors):
            img = sample[f'image_{sens}'].numpy()
            if self.as_time_series:
                # plot last time step
                img = img[-1, ...]
            if sens == 'S2':
                img = img[[2, 1, 0], ...]
                img = percentile_normalization(img.transpose(1, 2, 0))
            else:
                co_polarization = img[0]  # transmit == receive
                cross_polarization = img[1]  # transmit != receive
                ratio = co_polarization / cross_polarization

                # https://gis.stackexchange.com/a/400780/123758
                co_polarization = np.clip(co_polarization / 0.3, a_min=0, a_max=1)
                cross_polarization = np.clip(
                    cross_polarization / 0.05, a_min=0, a_max=1
                )
                ratio = np.clip(ratio / 25, a_min=0, a_max=1)

                img = np.stack((co_polarization, cross_polarization, ratio), axis=-1)

            axs[idx].imshow(img)
            axs[idx].axis('off')
            if show_titles:
                axs[idx].set_title(sens)

        if showing_predictions:
            pred = axs[ncols - 2].imshow(
                sample['prediction'].permute(1, 2, 0), cmap='YlGn'
            )
            plt.colorbar(pred, ax=axs[ncols - 2], fraction=0.046, pad=0.04)
            axs[ncols - 2].axis('off')
            if show_titles:
                axs[ncols - 2].set_title('Prediction')

        # plot target / only available in train set
        if 'label' in sample:
            target = axs[-1].imshow(sample['label'].permute(1, 2, 0), cmap='YlGn')
            plt.colorbar(target, ax=axs[-1], fraction=0.046, pad=0.04)
            axs[-1].axis('Off')
            if show_titles:
                axs[-1].set_title('Target')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SpaceNet 7 dataset."""

import os
from typing import ClassVar

from .base import SpaceNet


class SpaceNet7(SpaceNet):
    """SpaceNet 7: Multi-Temporal Urban Development Challenge.

    `SpaceNet 7 <https://spacenet.ai/sn7-challenge/>`_ is a dataset which
    consist of medium resolution (4.0m) satellite imagery mosaics acquired from
    Planet Labs' Dove constellation between 2017 and 2020. It includes ≈ 24
    images (one per month) covering > 100 unique geographies, and comprises >
    40,000 km2 of imagery and exhaustive polygon labels of building footprints
    therein, totaling over 11M individual annotations.

    Dataset features:

    * No. of train samples: 1423
    * No. of test samples: 466
    * No. of building footprints: 11,080,000
    * Area Coverage: 41,000 sq km
    * Chip size: 1024 x 1024
    * GSD: ~4m

    Dataset format:

    * Imagery - Planet Dove GeoTIFF

        * mosaic.tif

    * Labels - GeoJSON

        * labels.geojson

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2102.04420

    .. versionadded:: 0.2
    """

    directory_glob = os.path.join('**', '{product}')
    mask_glob = '*_Buildings.geojson'
    file_regex = r'global_monthly_(\d+.*\d+)'
    dataset_id = 'SN7_buildings'
    tarballs: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {0: ['SN7_buildings_train.tar.gz']},
        'test': {0: ['SN7_buildings_test_public.tar.gz']},
    }
    md5s: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {0: ['6eda13b9c28f6f5cdf00a7e8e218c1b1']},
        'test': {0: ['b3bde95a0f8f32f3bfeba49464b9bc97']},
    }
    valid_aois: ClassVar[dict[str, list[int]]] = {'train': [0], 'test': [0]}
    valid_images: ClassVar[dict[str, list[str]]] = {
        'train': ['images', 'images_masked'],
        'test': ['images_masked'],
    }
    valid_masks = ('labels', 'labels_match', 'labels_match_pix')
    chip_size: ClassVar[dict[str, tuple[int, int]]] = {
        'images': (1024, 1024),
        'images_masked': (1024, 1024),
    }
